from __future__ import annotations
from dataclasses import dataclass, field
import typing

import bson
from bstk_notitia.lib.abc.observer import ObserverABC
from datetime import datetime, UTC

if typing.TYPE_CHECKING:
    from modules.driver.mongodb import NotitiaModule as MongoDBDriver, DriverABC

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "mongodb",
        "required": True,
    }
}

# Establish and monitor a changestream, starting "now", and yield matching events as they arrive.
# The resume token is maintained within the observer
#
# When using this observer from an investigator, ensure you carefully consider the implications on the shift runner.
# A quiet collection that does not receive updates very often can cause the entire shift to run into overtime.
# Split your investigators into long lived pools until we get to py3.12
# ( https://docs.python.org/3.12/library/asyncio-task.html#waiting-primitives )


@dataclass
class NotitiaModule(ObserverABC):
    name = "Internal MongoDB Changestream"
    key = "internal/mongodb_changestream"
    resume_token: typing.AnyStr = field(init=False, default=None)

    if typing.TYPE_CHECKING:
        driver: typing.Dict[str, typing.Union[MongoDBDriver, DriverABC]]

    async def glance(self, document: typing.Dict) -> typing.Union[None, typing.Dict]:
        return document.get("fullDocument", None)

    async def monitor(
        self,
        collection: typing.AnyStr,
        lookup: typing.Optional[typing.Dict] = None,
        options: typing.Optional[typing.Dict] = None,
    ) -> typing.AsyncGenerator[bson.RawBSONDocument, None]:
        _pipeline = []
        if lookup:
            _pipeline.append(lookup)

        _collection = self.driver["mongodb"].get_collection(collection)
        if not self.resume_token and options and "resume_token" not in options:
            _cursor_start_time = self.driver["mongodb"].timestamp(datetime.now(UTC))
            options["start_at_operation_time"] = _cursor_start_time

        change_stream = self.driver["mongodb"].change_stream(
            collection=_collection,
            pipeline=_pipeline,
            **options,
        )
        async for token, change in self.driver["mongodb"].watch(change_stream):
            self.resume_token = token
            if not change:
                return

            doc = await self.glance(change[1])
            if not doc:
                continue

            yield doc



from .notitia import Notitia, Loader
from .error import (
    NotitiaError,
    NotitiaInvalidDriverException,
    NotitiaInvalidCasePayloadException,
    NotitiaPathExtensionWithoutSignature,
    NotitiaPathExtensionMissingSignature,
    NotitiaPathExtensionUnreferencedFile,
    NotitiaPathExtensionFileSignatureInvalid,
    NotitiaUnknownModule,
    NotitiaMalformedModulePath,
)

__name__ = "bstk_notitia"
__package__ = __name__
__all__ = [
    "Notitia",
    "NotitiaError",
    "NotitiaInvalidDriverException",
    "NotitiaInvalidCasePayloadException",
    "NotitiaPathExtensionWithoutSignature",
    "NotitiaPathExtensionMissingSignature",
    "NotitiaPathExtensionUnreferencedFile",
    "NotitiaPathExtensionFileSignatureInvalid",
    "NotitiaUnknownModule",
    "NotitiaMalformedModulePath",
]

import os

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
MODULE_PATHS = [
    "modules",
]
MODULE_TYPES = [
    "driver",
    "informant",
    "investigator",
    "observer",
    "reporter",
]

Notitia._module_loader = Loader(MODULE_DIR, MODULE_PATHS, MODULE_TYPES)



from __future__ import annotations
from dataclasses import dataclass, field
import typing
from datetime import datetime, UTC

from pymongo.errors import PyMongoError

from bstk_notitia.lib.abc.observer import ObserverABC

if typing.TYPE_CHECKING:
    from bstk_notitia.modules.driver.mongodb import (
        NotitiaModule as MongoDBDriver,
        DriverABC,
    )

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "mongodb",
        "required": True,
    },
}

# Where the mongo record observer does a "find all", and the mongo change stream observer "watches" for
# changes, the notitia collection stream observer wraps both up into a consolidated "find & watch" observer,
# whereby it will establish a change stream starting "now", find all pre-existing items matching the lookup,
# yield until empty, then move on to results from the change stream.
#
# When using this observer from an investigator, ensure you carefully consider the implications on the shift runner.
# A quiet collection that does not receive updates very often can cause the entire shift to run into overtime.
# Split your investigators into long lived pools until we get to py3.12
# ( https://docs.python.org/3.12/library/asyncio-task.html#waiting-primitives )


@dataclass
class NotitiaModule(ObserverABC):
    name = "Internal Mongo Collection Observer"
    key = "internal/mongo_collectionstream"

    if typing.TYPE_CHECKING:
        driver: typing.Dict[str, typing.Union[MongoDBDriver, DriverABC]]

    collection: typing.AnyStr = field(kw_only=True, default=None)

    lookup: typing.Optional[typing.Union[typing.List[typing.Dict], typing.Dict]] = (
        field(kw_only=True, default=None)
    )
    operations: typing.Optional[typing.List[typing.AnyStr]] = field(
        kw_only=True, default_factory=list
    )

    def __post_init__(self):
        if not self.operations:
            self.operations = ["insert", "update", "replace"]

    async def glance(self, document: typing.Dict) -> typing.Union[None, typing.Dict]:
        return document.get("fullDocument", None)

    async def monitor(
        self,
        lookup: typing.Optional[
            typing.Union[typing.List[typing.Dict], typing.Dict]
        ] = None,
        with_timeout: typing.Optional[int] = None,
    ) -> typing.AsyncGenerator[typing.Dict, None]:
        if not self.operations:
            return

        if self.lookup:
            if not lookup:
                lookup = self.lookup
            else:
                lookup = {**lookup, **self.lookup}

        _lookup = {"$match": {"operationType": {"$in": self.operations}}}
        if lookup:
            for _k, _v in lookup.items():
                _lookup["$match"][f"fullDocument.{_k}"] = _v

        prematch_hits = []

        _cursor_start_time = self.driver["mongodb"].timestamp(datetime.now(UTC))
        _collection = self.driver["mongodb"].get_collection(self.collection)
        async for doc in _collection.find(lookup):
            prematch_hits.append(doc["_id"])
            yield doc

        _stream_params = {
            "collection": _collection,
            "pipeline": [_lookup],
            "resume_token": None,
            "full_document": "updateLookup",
        }
        change_stream = self.driver["mongodb"].change_stream(
            **_stream_params,
            start_at_operation_time=_cursor_start_time,
        )

        try:
            async for token, change in self.driver["mongodb"].watch(
                changestream=change_stream, with_timeout=with_timeout, **_stream_params
            ):
                if not change and with_timeout is None:
                    # Something's gone wrong, stop.
                    break

                _stream_params["resume_token"] = token
                if not change and with_timeout is not None:
                    continue

                _t = change.get("operationType")
                if _t not in self.operations:
                    # We should never see this, so if we do - it's a sign things have gone poorly
                    break

                doc = await self.glance(change)
                if not doc:
                    continue

                if doc["_id"] in prematch_hits:
                    prematch_hits.remove(doc["_id"])
                    continue

                prematch_hits = []
                yield doc

        except PyMongoError as exc:
            if exc.timeout:
                yield None
            else:
                raise exc



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import typing

if typing.TYPE_CHECKING:
    from lib.abc.driver import DriverABC

"""
Reporters disseminate information provided to them by
investigators. The reporter understands how to publish to
a particular medium and formats the information to suit
its target audience.

Given the nature of data - reporters can be used to alter
the current state of information, (by writing to a source),
meaning they can make changes without necessarily making
information public.

Outside of invalid data or format limitations, reporters
cannot choose to withold information - they are impartial
intermediatries.
"""

NOTITIA_DEPENDENCIES: typing.Dict


@dataclass
class ReporterABC(ABC):
    department: typing.AnyStr = field(kw_only=True, default=None)
    name: typing.AnyStr = field(init=False)
    key: typing.AnyStr = field(init=False)
    type: typing.AnyStr = field(init=False, default="reporter")

    driver: typing.Dict[typing.AnyStr, DriverABC] = field(
        init=True, kw_only=True, default=None
    )

    def sign_on(self) -> typing.Dict:
        pass

    @abstractmethod
    async def receive(self, **kwargs):
        pass

    @abstractmethod
    async def publish(self):
        pass



from __future__ import annotations
from dataclasses import dataclass, field
import typing

from bstk_notitia.lib.abc.observer import ObserverABC

if typing.TYPE_CHECKING:
    from modules.driver.mongodb import NotitiaModule as MongoDBDriver, DriverABC

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "mongodb",
        "required": True,
    }
}


@dataclass
class NotitiaModule(ObserverABC):
    name = "Internal MongoDB Record"
    key = "internal/mongo_record"

    if typing.TYPE_CHECKING:
        driver: typing.Dict[str, typing.Union[MongoDBDriver, DriverABC]]

    collection: typing.AnyStr = field(kw_only=True, default=None)
    lookup: typing.Optional[typing.Union[typing.List[typing.Dict], typing.Dict]] = (
        field(kw_only=True, default=None)
    )

    async def glance(self, document: typing.Dict) -> typing.Union[None, typing.Dict]:
        pass

    async def monitor(
        self,
        options: typing.Optional[typing.Dict] = None,
        lookup: typing.Optional[
            typing.Union[typing.List[typing.Dict], typing.Dict]
        ] = None,
    ) -> typing.AsyncGenerator[typing.Dict, None]:
        if self.lookup:
            if not lookup:
                lookup = self.lookup
            else:
                lookup = {**lookup, **self.lookup}

        async for _doc in (
            self.driver["mongodb"].get_collection(self.collection).find(lookup)
        ):
            yield _doc



from datetime import datetime, UTC
from bstk_notitia.lib.abc.informant import InformantABC

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "null",
        "required": True,
    },
}


class NotitiaModule(InformantABC):
    name = "Internal HTTP Informant"
    key = "internal/http"

    async def enquire(self):
        return datetime.now(tz=UTC)

    async def inform(self):
        pass



from dataclasses import dataclass, field
import typing
from bstk_notitia.lib.abc.reporter import ReporterABC

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "null",
        "required": True,
    },
}


@dataclass
class NotitiaModule(ReporterABC):
    name = "Internal Echo Reporter"
    key = "internal/echo"

    timezone: typing.AnyStr = field(kw_only=True, default=None)

    async def receive(self, *args, **kwargs):
        if args:
            for _v in args:
                await self.publish(f"{_v}")

        if kwargs:
            for _k, _v in args:
                await self.publish(f"{_k}: {_v}")

    async def publish(self, data: str) -> typing.Any:
        print(data)



from dataclasses import dataclass, field
import datetime
import typing
from bstk_notitia.lib.abc.investigator import InvestigatorABC

"""
This sample investigator is a clock watcher.

It doesn't do anything interesting, but it does it reliably
from the moment it clocks on, until it clocks off.
"""

NOTITIA_DEPENDENCIES = {
    "driver": [
        {
            "type": "null",
            "required": True,
        },
        {
            "type": "null",
            "required": True,
        },
    ],
    "observer": {
        "type": "internal/tick",
        "required": True,
        "params": {"interval": "2"},
    },
    "informant": {
        "type": "internal/datetime",
        "required": True,
    },
    "reporter": {
        "type": "internal/echo",
        "required": True,
    },
}


@dataclass
class NotitiaModule(InvestigatorABC):
    name = "Internal Clock Investigator"
    key = "internal/clock"

    target_timezone: typing.List[typing.AnyStr] = field(
        kw_only=True, default_factory=list
    )

    async def start(self, **kwargs):
        if not self.observer or not self.informant or not self.reporter:
            return

        async for _ in self.observer["internal/tick"].monitor():
            data: datetime.datetime = await self.informant[
                "internal/datetime"
            ].enquire()
            await self.reporter["internal/echo"].receive(
                {"The date is": data.isoformat()}
            )



from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime, UTC
from decimal import Decimal, ROUND_HALF_UP
import typing
from uuid import uuid4
import simplejson

from bstk_notitia.lib.abc.reporter import ReporterABC

if typing.TYPE_CHECKING:
    from bstk_notitia.modules.driver.mongodb import (
        NotitiaModule as MongoDBDriver,
        DriverABC,
    )

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "mongodb",
        "required": True,
    },
}


@dataclass
class NotitiaModule(ReporterABC):
    name = "Internal Notitia Case Reporter"
    key = "internal/notitia_case"

    if typing.TYPE_CHECKING:
        driver: typing.Dict[str, typing.Union[MongoDBDriver, DriverABC]]

    collection: typing.AnyStr = field(kw_only=True, default=None)

    async def receive(
        self,
        action: str,
        record: typing.Dict,
        data: typing.Optional[
            typing.Union[typing.List[typing.Dict], typing.Dict]
        ] = None,
    ) -> typing.Union[bool, str, typing.List[str]]:
        if action not in ("clone", "new"):
            return False
        _cases = []

        if data is None:
            data = {"number": None}

        if not isinstance(data, list):
            data = [data]

        for _entry in data:
            _entry["number"] = str(uuid4())
            _entry["assigned_investigator"] = None
            _entry["created_at"] = datetime.now(tz=UTC)
            _entry["updated_at"] = None
            _entry["closed_at"] = None

            if action == "clone":
                _entry["log"] = [
                    {
                        "type": "clone",
                        "module": "notitia",
                        "date": _entry["created_at"],
                    }
                ]
            else:
                _entry["log"] = [
                    {
                        "type": "created",
                        "module": "notitia",
                        "date": _entry["created_at"],
                    }
                ]

            _new_case = {**simplejson.loads(self.json_encode(record)), **_entry}
            if "_id" in _new_case:
                del _new_case["_id"]

            print(f"{self.key} created {action} case {_new_case['number']}")
            _res = await self.publish(_new_case)
            if not _res:
                continue
            _cases.append(_res)

        if not len(_cases):
            return False

        return _cases if len(_cases) > 1 else _cases[0]

    async def publish(self, case_data: typing.Dict):
        res = (
            await self.driver["mongodb"]
            .get_collection(self.collection)
            .insert_one(document=case_data)
        )

        if hasattr(res, "inserted_id") and res.inserted_id:
            return res.inserted_id

        return False

    def json_encode(self, value: typing.Any) -> str:
        """
        A consistency layer around json.dumps.
        (you usually just get the .__str__() output)
        Handles:
            * Date/Datetime
            * UUID
            * Decimals
        Do _not_ use this if you're targeting mongodb, use the BSON methods instead.
        """
        return simplejson.dumps(value, default=self.json_convert, use_decimal=True)

    def json_convert(self, elem: typing.Any) -> str:
        if isinstance(elem, object):
            if hasattr(elem, "_to_json"):
                return elem._to_json()
            if hasattr(elem, "isoformat"):
                return elem.isoformat()
            if hasattr(elem, "strftime"):
                return elem.strftime("%Y-%m-%d")
            if hasattr(elem, "to_decimal"):
                elem = elem.to_decimal()
            if hasattr(elem, "quantize"):
                return float(elem.quantize(Decimal("1.0000"), rounding=ROUND_HALF_UP))
            if hasattr(elem, "value"):
                return elem.value

        try:
            return str(elem)
        except Exception:
            return None



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import typing

if typing.TYPE_CHECKING:
    from lib.abc.driver import DriverABC

"""
Observers are the initial source for all investigators. They
are unbiased and focus only on validating that the subject
information matches a fixed criteria.

An investigator will give an observer guidelines regarding
what kind of information it cares about, and the observer
will provide that back to the investigator once it sees
matching information.

Observers only exist when an investigator requests one and are
the only mechanism by which the system will start processing
information.

Investigators are inherently lazy, so if the observer reports
nothing, then the investigator will happily sleep forever.

If an observer relies on external information, (ie: queue entry,
data changes, etc), it needs to be aware that it is not the only
observer in the system, so it should track its own state, ensuring
that it doesn't report invalid or duplicate information, nor
prevent other observers from working with the source data.
"""

NOTITIA_DEPENDENCIES: typing.Dict


@dataclass
class ObserverABC(ABC):
    department: typing.AnyStr = field(kw_only=True, default=None)
    name: typing.AnyStr = field(init=False)
    key: typing.AnyStr = field(init=False)
    type: typing.AnyStr = field(init=False, default="observer")

    driver: typing.Dict[typing.AnyStr, DriverABC] = field(
        init=True, kw_only=True, default=None
    )

    def sign_on(self) -> typing.Dict:
        pass

    @abstractmethod
    async def glance(self, *args, **kwargs) -> bool:
        """Quickly review the input data and see whether this observe should be looking at the data"""
        pass

    @abstractmethod
    async def monitor(
        self, *args, **kwargs
    ) -> typing.Union[typing.Any, typing.AsyncGenerator[typing.Any, typing.Any]]:
        """With the given args, keep watch until there is something to report back with"""
        pass



from dataclasses import dataclass, field
import typing
from bstk_notitia.lib.abc.observer import ObserverABC
from timedelta_isoformat import timedelta
from datetime import datetime
from asyncio import sleep

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "null",
        "required": True,
    },
}


@dataclass
class NotitiaModule(ObserverABC):
    name = "Internal Tick Observer"
    key = "internal/tick"
    interval: typing.SupportsFloat = field(kw_only=True, default=2)

    async def glance(self, target: typing.AnyStr) -> bool:
        if not target:
            return False

        try:
            timedelta.fromisoformat(target)
        except Exception:
            return False

        return True

    async def monitor(
        self, target: typing.Optional[float] = None
    ) -> typing.Union[datetime, typing.AsyncGenerator[datetime, None]]:
        _is_persistent: bool = False
        if target:
            _is_persistent = True

        if not target and self.interval:
            target = self.interval

        target = f"PT{target}S"

        if not await self.glance(target):
            raise ValueError("Invalid tick target")

        delta = timedelta.fromisoformat(target)
        now = datetime.now()

        while 1:
            until = now + delta
            tosleep = (until - now).total_seconds()
            await sleep(tosleep)
            yield until
            if not _is_persistent:
                break



import inspect
import typing
from uuid import uuid4

from bstk_notitia.lib.abc.driver import DriverABC
from bstk_notitia.error import (
    NotitiaInvalidDriverParameter,
    NotitiaUnsupportedDriverCall,
)


class NotitiaModule(DriverABC):
    name = "Passthrough driver"
    key = "driver/passthrough"
    connections: typing.Dict[str, typing.Any]

    #
    # The passthrough driver provides a consistent interface for custom drivers.
    # You can use it as a simple registry to provide functionality to your modules without
    # having to create a full blown driver.
    #
    # The "dsn" provided to connect will be the object the driver will use, we generate
    # and return a "reference", which can be used to later disconnect or make calls.
    # You can make a "call" by providing the connection reference along with the method / params to execute.
    # Because we don't have a proper dsn, calling connect multiple times with the same
    # client will generate multiple references - so don't do that.
    #
    # Note that passthrough / custom drivers are not licenses to do random things in random ways.
    # Notitia modules exist to provide consistent and predictable behaviour, i.e. don't do weird shit.
    #
    def connect(self, dsn: typing.Any) -> typing.Any:
        if not dsn or not hasattr(dsn, "__class__"):
            raise NotitiaInvalidDriverParameter("Invalid DSN")

        _ref = self._genref()
        self.connections[_ref] = dsn

        return _ref

    # Remember - the dsn here is the reference returned by .connect
    def disconnect(self, dsn: str):
        if not dsn:
            return

        if dsn not in self.connections:
            return

        del self.connections[dsn]
        return

    def _genref(self) -> str:
        return str(uuid4())

    async def call(
        self, dsn: str, method: str, *args, **kwargs
    ) -> typing.Optional[typing.Dict]:
        if not dsn or dsn not in self.connections:
            raise NotitiaInvalidDriverParameter("Invalid DSN")

        _conn = self.connections[dsn]
        _call = getattr(_conn, method, None)
        if (
            _call is None
            or not callable(_call)
            or not inspect.iscoroutinefunction(_call)
        ):
            raise NotitiaUnsupportedDriverCall(
                f"{method} not suitable to be called on {_conn} (not defined, not callable, or not coro)"
            )

        return await _call(*args, **kwargs)



from __future__ import annotations
from dataclasses import dataclass, field
from types import NoneType
import typing
from bstk_notitia.lib.abc.reporter import ReporterABC

if typing.TYPE_CHECKING:
    from modules.driver.mongodb import NotitiaModule as MongoDBDriver, DriverABC
    from pymongo.results import (
        UpdateResult,
        InsertManyResult,
        InsertOneResult,
        DeleteResult,
    )
    from asyncio import Future

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "mongodb",
        "required": True,
    },
}


@dataclass
class NotitiaModule(ReporterABC):
    name = "Internal MongoDB Record Reporter"
    key = "internal/mongodb_record"

    collection: typing.Optional[typing.AnyStr] = field(kw_only=True, default=None)

    if typing.TYPE_CHECKING:
        driver: typing.Dict[str, typing.Union[MongoDBDriver, DriverABC]]

    async def receive(
        self,
        action: str,
        data: typing.Union[typing.List[typing.Dict], typing.Dict] = None,
        record: typing.Optional[typing.Dict] = None,
        constraints: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ):
        brief = None
        if action == "update":
            if not record or not record.get("_id"):
                return False

            if data is None:
                raise ValueError("data is required for update")

            stream = self.driver["mongodb"].get_collection(self.collection).update_one
            if not constraints:
                constraints = {"_id": record["_id"]}
            else:
                constraints["_id"] = record["_id"]

            if not self._has_update_operators(data):
                data = {"$set": data}

            data["$currentDate"] = {"updated_at": True}

            stream_params = {"filter": constraints, "update": data}

        elif action == "insert":
            if data is None:
                raise ValueError("data is required for insert")

            if isinstance(data, list):
                stream = (
                    self.driver["mongodb"].get_collection(self.collection).insert_many
                )
                for _d in data:
                    _d["$currentDate"] = {"created_at": True}

                stream_params = {"documents": data}
            else:

                data["$currentDate"] = {"created_at": True}

                stream = (
                    self.driver["mongodb"].get_collection(self.collection).insert_one
                )
                stream_params = {"document": data}

        elif action == "apply":
            if data is None:
                raise ValueError("data is required for apply")

            if not constraints:
                return False

            data["$currentDate"] = {"updated_at": True}

            stream = self.driver["mongodb"].get_collection(self.collection).update_many
            stream_params = {"filter": constraints, "update": data}

        elif action == "remove":
            if not constraints:
                return False

            stream = self.driver["mongodb"].get_collection(self.collection).delete_many
            stream_params = {"filter": constraints}

        if not stream:
            return False

        brief = stream(**stream_params)

        if not brief:
            return False

        res = await self.publish(brief)
        return res

    def _has_update_operators(self, data: typing.Dict[str, any]) -> bool:
        _k = list(data)[0]
        if _k[0:1] == "$":
            return True

        return False

    async def publish(self, brief: Future) -> typing.Any:
        result: typing.Union[
            UpdateResult, InsertManyResult, InsertOneResult, DeleteResult
        ] = await brief

        if not result.acknowledged:
            return False

        if hasattr(result, "matched_count") and not result.matched_count:
            return NoneType

        if hasattr(result, "inserted_ids") and result.inserted_ids:
            return result.inserted_ids

        if hasattr(result, "inserted_id") and result.inserted_id:
            return result.inserted_id

        if hasattr(result, "modified_count") and result.modified_count:
            return result.modified_count

        if hasattr(result, "deleted_count") and result.deleted_count:
            return result.deleted_count



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import typing

"""
A Notitia driver is a storage or streaming data source that can be used
across all other modules.

Unlike other modules, drivers are instanciated once and made available to
other modules that request it in their dependency spec.

Notitia holds and provides connection strings, authentication information,
etc, so modules purely need to know what dsn they should use.
"""


@dataclass
class DriverABC(ABC):
    department: typing.AnyStr = field(kw_only=True, default=None)
    name: typing.AnyStr = field(init=False)
    key: typing.AnyStr = field(init=False)
    type: typing.AnyStr = field(init=False, default="driver")

    connections: typing.Dict[typing.AnyStr, typing.Any] = field(
        init=False, default_factory=dict
    )

    def set_up(self, dsn: typing.Any) -> typing.Dict:
        self.connect(dsn)

    @abstractmethod
    def connect(self, dsn: typing.Any):
        """
        Connect to the specified DSN and return a "client like" reference that
        can be used to communicate with the service
        """
        pass

    @abstractmethod
    def disconnect(self, dsn: typing.Optional[typing.Any]):
        """
        If provided with a dsn and we have an active connection using it, disconnect
        from it.
        If no dsn is provided, disconnect from everything (shut down)
        """
        pass



from __future__ import annotations
import asyncio
from dataclasses import dataclass, field
from enum import Enum
import typing
import queue
from uuid import uuid4

import concurrent.futures
import sys
import traceback


if typing.TYPE_CHECKING:
    from lib.abc.investigator import InvestigatorABC

"""
The roster is responsible for managing investigators and their
associated tasks.

When an investigator is added to the roster, it will:
 - Create a new threadpool for that investigator and its observers
 - Instantiate any observers required by the investigator
 - Connect a dispatcher to the investigators threadpool
 - Ensure investigators clean up after themselves

The roster limits the number of active investigators clocked on,
but can have an infinite pool of investigators waiting for a shift.

It also controls how many jobs can be executed in a shift, along with
how long an investigator can run for before it needs to take a break.

If an investigator is marked as "always on", the shift limit still
applies, but the investigator will be immediately reinstated as soon
as a slot becomes available, (always on investigators have a priority of
1, the rest as priority 10). Investigators that are not "always on"
form an orderly queue, based on when they were added to the roster,
(FIFO).

So long as an investigator is not in the middle of anything (ie:
waiting for an observer), it will be clocked off. Slow investigations
lead to slower shift cycles and limit the overall throughput

"""


class RosterType(Enum):
    PERMANENT = 1
    CASUAL = 2


class Roster:
    investigator_limit: int = 10
    shift_limit: int = 30
    _shift_pools: typing.Dict[str, queue.Queue] = None
    _active_shift: ShiftManager = None
    debug: bool = False

    def __init__(
        self, limit: typing.Optional[int] = None, duration: typing.Optional[int] = None
    ) -> None:
        if limit:
            self.investigator_limit = limit
        if duration:
            self.set_duration(duration)

    def set_duration(self, duration: int) -> None:
        self.shift_limit = duration

    def _initialise_shift_pools(self):
        if self._shift_pools is not None:
            return

        self._shift_pools = {
            RosterType.PERMANENT.name: queue.Queue(),
            RosterType.CASUAL.name: queue.Queue(),
        }

    def add(
        self,
        investigator: InvestigatorABC,
        type: RosterType,
        startup_params: typing.Optional[typing.Dict] = None,
    ):
        if not self._shift_pools:
            self._initialise_shift_pools()

        if not isinstance(type, RosterType):
            raise ValueError("`type` must be a `RosterType`")

        self.debug and print(
            f"Pushing investigator {investigator.key}({investigator.name}) into {type.name} shift pool"
        )
        self._shift_pools[type.name].put_nowait(investigator)

    def get_shift(self) -> ShiftManager:
        if not self._shift_pools:
            raise Exception("No shift pools available - add an investigator first")

        if self._active_shift is not None:
            raise Exception("Already an active shift running - end it first")

        self._active_shift = ShiftManager(
            limit=self.investigator_limit,
            duration=self.shift_limit,
        )

        self._active_shift.populate_pools(self._shift_pools)

        return self._active_shift

    async def end_shift(self) -> None:
        if self._active_shift is None:
            return

        await self._active_shift.end()
        self.debug and await self._active_shift._shift_coordinator.stats()


class ShiftManager:
    """
    The shift manager sets up investigations, ready for the shift
    coordinator to distribute to workstations.

    Investigations that are "permanent" have a higher priority than
    casual investigations.

    The default cycle rate is set to `2` (a cycle ever two seconds).

    Once pools are populated, `start()` will return a coro that
    can be thrown over to an executor and run
    """

    pools: typing.Dict[str, queue.Queue[Investigation]] = None
    _shift_coordinator: ShiftCoordinator = None
    debug: bool = False

    def __init__(
        self,
        limit: typing.Optional[int] = 10,
        duration: typing.Optional[int] = 0,
    ):
        cycle_rate = 2
        cycle_limit = None
        if duration:
            cycle_limit = duration / cycle_rate

        self._shift_coordinator = ShiftCoordinator(
            limit, cycle_rate=cycle_rate, cycle_limit=cycle_limit
        )

    def populate_pools(self, pools: typing.Dict[str, queue.Queue]):
        for _rt in (RosterType.PERMANENT, RosterType.CASUAL):
            while not pools[_rt.name].empty():
                _investigation = Investigation(
                    priority=_rt.value,
                    id=str(uuid4()),
                    investigator=pools[_rt.name].get(),
                )

                self.debug and print(
                    f"Added investigation #{_investigation.id}[{_investigation.investigator.key}]"
                    + f" with {_rt.name} priority"
                )
                self._shift_coordinator.add_investigation(_investigation)

    """
    Changing the cycle rate and limit here imply you know what you
    are trying to do. Otherwise, just use `Roster.set_duration` to
    change how long a shift lasts for prior to getting the shift.
    """

    def start(
        self,
        cycle_rate: typing.Optional[float] = None,
        cycle_limit: typing.Optional[float] = None,
        workstation_isolation: typing.Optional[bool] = None,
        finish_callback: typing.Optional[typing.Callable] = None,
    ) -> typing.Coroutine:
        return self._shift_coordinator.start(
            cycle_rate=cycle_rate,
            cycle_limit=cycle_limit,
            workstation_isolation=workstation_isolation,
            finish_callback=finish_callback,
        )

    def end(self) -> typing.Coroutine:
        return self._shift_coordinator.finish()


class ShiftCoordinator:
    """
    The shift coordinator creates a pool of workstations that
    can accept investigations. The workstations collect
    investigations from the shift pool based on their priority
    and will deposit them back onto the queue once each
    investigation is complete

    The shift coordinator ensures the requested number
    of workstations are available to be used for investigation
    along with how long the shift will go for.

    Calling `start()` will return a coroutine which can be
    executed in the appropriate context.
    """

    investigations: typing.List[Investigation] = None
    debug: bool = False

    _started: bool = False
    _finishing: bool = False

    _tickrate: float
    _ticklimit: float

    shift_pool: queue.PriorityQueue[Investigation] = None
    workstations: typing.List[Workstation] = None
    workstation_isolation: bool

    max_workstation_count: int = 10
    occupied_workstations: set

    def __init__(
        self,
        limit: typing.Optional[int] = 10,
        investigations: typing.Optional[typing.List[Investigation]] = None,
        cycle_rate: typing.Optional[float] = 1,
        cycle_limit: typing.Optional[float] = 0,
        workstation_isolation: typing.Optional[bool] = False,
    ) -> None:
        self.investigations = []
        if investigations:
            for _investigation in investigations:
                self.add_investigation(_investigation)

        if limit < 1:
            raise ValueError("`limit` must be a non-zero integer")
        self.max_workstation_count = limit

        if cycle_rate:
            self._tickrate = cycle_rate

        if cycle_limit:
            self._ticklimit = cycle_limit

        self.workstation_isolation = workstation_isolation

        self.shift_pool = asyncio.PriorityQueue()
        self.occupied_workstations = set()

    @property
    def occupied_workstation_count(self) -> int:
        return len(self.occupied_workstations)

    def open_workstation(self, workstation: Workstation) -> None:
        open = workstation.open_for_investigation()
        self.occupied_workstations.add(open)
        asyncio.create_task(self._open_workstation(open))

    async def _open_workstation(self, coro: typing.Coroutine) -> None:
        try:
            await coro
        except Exception:
            print("Exception when opening workstation..")
            print("-" * 60)
            traceback.print_exc(file=sys.stdout, chain=True)
            print("-" * 60)
        finally:
            self.occupied_workstations.remove(coro)

    def add_investigation(self, investigation: Investigation) -> None:
        self.investigations.append(investigation)

    def start(
        self,
        cycle_rate: typing.Optional[float] = None,
        cycle_limit: typing.Optional[float] = None,
        workstation_isolation: typing.Optional[bool] = None,
        finish_callback: typing.Optional[typing.Callable] = None,
    ) -> typing.Coroutine:
        if self._started:
            raise RuntimeError("Already started")

        if len(self.investigations) < 1:
            raise RuntimeError("Nothing to do")

        if cycle_rate is not None:
            self._tickrate = cycle_rate

        if cycle_limit is not None:
            self._ticklimit = cycle_limit

        if workstation_isolation is not None:
            self.workstation_isolation = workstation_isolation

        self.max_workstation_count = min(
            self.max_workstation_count, len(self.investigations)
        )

        for _i in self.investigations:
            self.debug and print(
                f"Adding {_i.id}:{_i.investigator.key} into the shift pool"
            )
            self.shift_pool.put_nowait(_i)

        self.workstations = [
            Workstation(
                id=_, queue=self.shift_pool, isolation=self.workstation_isolation
            )
            for _ in range(self.max_workstation_count)
        ]

        return self._start(finished_callback=finish_callback)

    async def _start(
        self, finished_callback: typing.Optional[typing.Callable] = None
    ) -> None:
        self._started = True
        _ticker = 0

        # print(f"< [^{self._ticklimit}] @ [~{self._tickrate}] >")
        while self._finishing is not True:
            if _ticker > 0:
                await asyncio.sleep(self._tickrate)

            _ticker += 1
            if self._ticklimit and self._ticklimit < _ticker:
                await self.finish()
                # print(f"< {_ticker} X {self._ticklimit} >")
                break

            # No point even trying if the pool is full
            if self.shift_pool.full():
                # print(f"<!- {_ticker} -!>")
                continue

            idle_workstations = [_w for _w in self.workstations if _w.available is True]
            if not idle_workstations:
                # print(f"<o {_ticker} o>")
                continue

            for idle_workstation in idle_workstations:
                if not idle_workstation:
                    # print(f"<? {_ticker} ?>")
                    continue

                idle_workstation.available = False
                self.open_workstation(idle_workstation)
                # print(f"< {_ticker} +{idle_workstation.id} >")

        while self.occupied_workstation_count > 0:
            await self.finish()
            await asyncio.sleep(0.3)

        if finished_callback:
            await finished_callback()

    async def finish(self):
        """
        Prevent workstations from being opened and signal
        all active workstations that they should finish
        """
        self._finishing = True
        for _w in self.workstations:
            if _w.closing:
                continue
            await _w.close()

        return

    async def stats(self):
        print(f"Pool size {self.shift_pool.qsize()}")
        for _i in self.investigations:
            print(
                f"Investigator {_i.id}:{_i.investigator.key} cycle count: {_i.cycle_count}"
            )


"""
An investigation is a prioritisation and identification
concept - it does very little other than ensure correct
queueing behaviour
"""


@dataclass(order=True)
class Investigation:
    priority: int
    id: int = field(compare=False)
    investigator: InvestigatorABC = field(compare=False)
    cycle_count: int = field(init=False, default=0)


"""
Workstations are where investigations are worked on by
investigators. They are semi-autonomous, pull items off
the shift pool when available and put them back afterwards.

Workstations are well behaved, in that they will reboot
themselves if no investigations become available within
`idle_limit` seconds - helping to speed up shift closure.

Workstations have limited patience and will not wait for
and investigator if they're taking too long. To allow
for other investigations that may have faster or more urgent
work, it will wait for an investigator no longer than
`session_limit` seconds to complete their work.

Each workstation maintains its own event loop, ensuring
that the workstation remains nice and clean between
investigations and that a heavy workload on one
workstation does impact other workstations. To facilitate
this cleanliness, thread pools are only stood up when
and investigation starts and torn down as soon as it finishes.
"""


class Workstation:
    id: int
    queue: asyncio.PriorityQueue[Investigation]
    available: bool
    closing: bool
    idle_limit: int
    session_limit: int
    isolation: bool
    _debug: bool

    def __init__(
        self,
        id: int,
        queue: asyncio.PriorityQueue[Investigation],
        idle_limit: typing.Optional[int] = 10,
        session_limit: typing.Optional[int] = None,
        isolation: typing.Optional[bool] = False,
        debug: typing.Optional[bool] = False,
    ):
        self.available = True
        self.closing = False
        self.id = id
        self.queue = queue
        self.idle_limit = idle_limit
        self.session_limit = session_limit
        self.isolation = isolation
        self._debug = debug

    async def open_for_investigation(self):
        self.available = False
        try:
            async with asyncio.timeout(self.idle_limit):
                investigation = await self.queue.get()
        except TimeoutError:
            # print(f"< -{self.id} >")
            self.available = True
            return

        investigation.cycle_count += 1
        self._debug and print(
            f"Workstation #{self.id} occupied by"
            + f" [p{investigation.priority}]{investigation.id} /"
            + f" {investigation.investigator.key} {asyncio.current_task().get_name()} (#{investigation.cycle_count})",
        )

        activity = None
        try:
            _coro = investigation.investigator.start()
            _check_interval = self.session_limit or 0.5

            if self.isolation:
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=3, thread_name_prefix=f"nwk_{self.id}"
                ) as executor:
                    activity = executor.submit(asyncio.run, _coro)

                    while not activity.done():
                        try:
                            # print(f"#{self.id} waiting for {_check_interval}")
                            _ = await asyncio.wait_for(
                                asyncio.shield(activity), _check_interval
                            )
                            # print(f"#{self.id} complete..")
                            break
                        except asyncio.CancelledError as e:
                            self._debug and print(
                                f"#{self.id} investigator was cancelled.."
                            )
                            raise e
                        except TimeoutError as e:
                            if getattr(investigation.investigator, "busy", None):
                                # print(f"#{self.id} waiting - the investigator is busy..")
                                continue

                            if self.closing:
                                # print(f"#{self.id} closing out investigator")
                                activity.cancel()
                                raise e

                            continue

            else:
                _coro_task = asyncio.create_task(_coro)
                while True:
                    try:
                        # print(f"#{self.id} waiting for {_check_interval}")
                        _ = await asyncio.wait_for(
                            asyncio.shield(_coro_task), _check_interval
                        )
                        # print(f"#{self.id} complete..")
                        break
                    except asyncio.CancelledError as e:
                        self._debug and print(
                            f"#{self.id} investigator was cancelled.."
                        )
                        raise e
                    except TimeoutError as e:
                        if getattr(investigation.investigator, "busy", None):
                            # print(f"#{self.id} waiting - the investigator is busy..")
                            continue

                        if self.closing:
                            # print(f"#{self.id} closing out investigator")
                            _coro_task.cancel()
                            raise e

                        continue

            self._debug and print(
                f"Workstation #{self.id} relinquished by"
                + f" [p{investigation.priority}]{investigation.id} /"
                + f" {investigation.investigator.key} {asyncio.current_task().get_name()}",
            )

        except TimeoutError:
            if activity:
                activity.cancel()
            print(
                f"Workstation #{self.id} timed out"
                + f" [p{investigation.priority}]{investigation.id} /"
                + f" {investigation.investigator.key} {asyncio.current_task().get_name()}",
            )

        except Exception:
            print(
                f"Workstation #{self.id} caught an error in"
                + f" {investigation.id} / {investigation.investigator.key}"
            )
            traceback.print_exc(file=sys.stdout, chain=True)

        finally:
            self._debug and print(
                f"Returning {investigation.id} / {investigation.investigator.key}"
                + f"to the pool (cycles: {investigation.cycle_count})"
            )
            self.queue.task_done()
            await self.queue.put(investigation)
            # print(f"< -{self.id} >")
            self.available = True

    async def close(self):
        self._debug and print(f"Workstation #{self.id} being asked to close")
        self.closing = True
        while not self.available:
            await asyncio.sleep(0.1)

        return



from __future__ import annotations
import typing
from bstk_notitia.lib.abc.investigator import InvestigatorABC

if typing.TYPE_CHECKING:
    import datetime
    from modules.driver.mongodb import NotitiaModule as MongoDBDriver, DriverABC

"""
This sample investigator is a mongodb record "enricher".

It asks the observer to collect records from a
particular collection that have not been collected,
gets the current date from the date reporter
and tells the mongo_record reporter that it
has changes to report.

Being a busywork investigator, it always tells
the reporter that it has a new record to report
before the observer is even engaged - so its
always got _something_ to do.
"""

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "mongodb",
        "required": True,
        "params": {
            "dsn": "mongodb://localhost",
        },
    },
    "observer": {
        "type": "internal/mongo_record",
        "required": True,
        "params": {
            "lookup": {"collected": False},
        },
    },
    "informant": {
        "type": "internal/datetime",
        "required": True,
    },
    "reporter": {
        "type": "internal/mongo_record",
        "required": True,
    },
}


class NotitiaModule(InvestigatorABC):
    name = "Internal MongoDB Record Collector"
    key = "internal/mongo_record_processor"
    driver: typing.Dict[str, typing.Union[MongoDBDriver, DriverABC]]

    async def start(self, **kwargs):

        async for document in self.observer["internal/mongo_record"].monitor():
            data: datetime.datetime = await self.informant[
                "internal/datetime"
            ].enquire()
            res = await self.reporter["internal/mongo_record"].receive(
                action="update",
                record=document,
                data={"collected": True, "collected_at": data},
            )
            print(f"{res}")



import os
import re
from pathlib import Path
import importlib.util
from types import ModuleType
import typing

from .error import (
    NotitiaError,
    NotitiaInvalidModulePath,
    NotitiaMalformedModulePath,
    NotitiaMissingModule,
    NotitiaPathExtensionFileSignatureInvalid,
    NotitiaPathExtensionMissingSignature,
    NotitiaPathExtensionUnreferencedFile,
    NotitiaPathExtensionWithoutSignature,
    NotitiaUnknownModule,
)

NOTITIA_MODULE_DEPLIST = {
    "investigator": (
        "driver",
        "observer",
        "informant",
        "reporter",
    ),
    "observer": ("driver",),
    "informant": ("driver",),
    "reporter": ("driver",),
    "driver": (),
}


class Loader:
    module_paths: typing.Optional[typing.List[typing.AnyStr]] = None
    _processed_paths: typing.List[typing.AnyStr] = None
    _module_path_lookup: typing.Dict[typing.AnyStr, typing.AnyStr] = None

    modules: typing.Dict[typing.AnyStr, typing.List[typing.AnyStr]] = None
    _modules: typing.Dict[typing.AnyStr, typing.Dict[typing.AnyStr, ModuleType]] = None

    module_types: typing.List[typing.AnyStr] = None

    signing_key: typing.Optional[typing.AnyStr] = None
    untrusted_paths: typing.Optional[typing.List[typing.AnyStr]] = None

    def __init__(self, base_dir, module_paths=None, module_types=None):
        self.base_dir = Path(base_dir)
        if module_paths is None:
            return

        self.module_paths = module_paths
        self.untrusted_paths = []
        self._processed_paths = []

        self.modules = {}
        self._modules = {}
        self._module_path_lookup = {}

        if module_types is None:
            return

        self.module_types = module_types
        self._collect_modules_from_paths()

    def get_module_types(self):
        return list(self.modules.keys())

    def list_modules(self, type: str = None):
        if type is None:
            return self.modules

        return self.modules.get(type, False)

    def get_module(self, type: str, code: str):
        if type not in self._modules:
            return False

        return self._modules[type].get(code, False)

    def add_paths(self, paths: typing.List[typing.AnyStr]):
        for path in paths:
            path = Path(path)
            if not path.exists() or not path.is_dir():
                continue

            try:
                self.base_dir.parent.relative_to(path)
            except ValueError:
                if not self.signing_key:
                    raise NotitiaPathExtensionWithoutSignature(
                        "Refusing to include an external module path without a signature"
                    )
                self.untrusted_paths.append(path)
                continue

            self.module_paths.append(path)

        self._collect_modules_from_paths()

    def _collect_modules_from_paths(self):
        if self.module_paths is None:
            return False

        for module_path in [*self.module_paths, *self.untrusted_paths]:
            module_path = Path(
                os.path.join(self.base_dir, module_path)
                if module_path not in self.untrusted_paths
                else module_path
            )
            if (
                not module_path.exists()
                or not module_path.is_dir()
                or module_path in self._processed_paths
            ):
                continue

            module_files = list(module_path.glob("**/*.py"))
            if not len(module_files):
                continue

            if module_path in self.untrusted_paths:
                try:
                    self._validate_signature(module_path, module_files)
                except (
                    NotitiaPathExtensionWithoutSignature,
                    NotitiaPathExtensionMissingSignature,
                    NotitiaPathExtensionUnreferencedFile,
                    NotitiaPathExtensionFileSignatureInvalid,
                ) as e:
                    self.untrusted_paths.remove(module_path)
                    raise e

            self._collect_modules_from_path(module_path, module_files)

            self._processed_paths.append(module_path)

    def _collect_modules_from_path(
        self,
        module_path: Path,
        module_files: typing.List[Path],
    ):
        for module_file in module_files:
            _mod_path = module_file.relative_to(module_path)

            if len(_mod_path.parts) > 3 or len(_mod_path.parts) < 2:
                raise NotitiaMalformedModulePath()

            _mod_type = _mod_path.parts[0]
            if _mod_type not in self.module_types:
                continue

            if _mod_type not in self.modules:
                self.modules[_mod_type] = []

            module_name = re.sub(r"\.py$", "", module_file.name)
            if module_name[0:1] == "_":
                # It's a valid module that's passed signature checks - but it's not executable, so we don't load it
                continue

            if module_file.parent.name != _mod_type:
                module_name = f"{module_file.parent.name}/{module_name}"

            self.modules[_mod_type].append(module_name)
            self._module_path_lookup[f"{_mod_type}/{module_name}"] = module_file

    def load_module(self, module_type: typing.AnyStr, module_name: typing.AnyStr):
        if self.module_paths is None:
            return False

        if (
            module_type not in self.modules
            or module_name not in self.modules[module_type]
        ):
            raise NotitiaUnknownModule()

        if module_type not in self._modules:
            self._modules[module_type] = {}

        if module_name in self._modules[module_type]:
            return

        _mod_name = f"{module_type}/{module_name}"
        if _mod_name not in self._module_path_lookup:
            raise NotitiaMissingModule()

        module_file = self._module_path_lookup[_mod_name]

        if not os.path.exists(module_file):
            raise NotitiaInvalidModulePath()

        _mod = __class__._import_module(_mod_name, module_file)
        self._modules[module_type][module_name] = _mod
        if not _mod:
            return

        self._load_module_deps(module_type, _mod)

    def _load_module_deps(self, module_type: str, module: ModuleType):
        acceptable_deps = NOTITIA_MODULE_DEPLIST[module_type]

        _deps: typing.Dict[str, typing.Union[typing.List, typing.Dict]] = getattr(
            module, "NOTITIA_DEPENDENCIES", {}
        )
        for _type, _deplist in _deps.items():
            if _type not in acceptable_deps:
                raise NotitiaError(
                    f"{module_type} modules cannot depend on `{_type}` modules"
                )

            if isinstance(_deplist, dict):
                _deplist = [_deplist]

            for _dep in _deplist:
                try:
                    self.load_module(_type, _dep["type"])
                except NotitiaError as ex:
                    if "required" in _dep and _dep["required"]:
                        raise ex

    def _validate_signature(self, path: Path, files: typing.List[Path]):
        if not self.signing_key:
            raise NotitiaPathExtensionWithoutSignature(
                "Refusing to process - no signature provided"
            )
        sig_file = Path(path, ".verify")
        if not sig_file.exists():
            raise NotitiaPathExtensionMissingSignature(
                "Refusing to process - no .verify file found"
            )

        _signatures = {}
        with sig_file.open("r") as _f:
            for _l in _f.readlines():
                _l = _l.strip().split(":", 1)
                _signatures[_l[0]] = _l[1]

        import hashlib
        import hmac

        for _file in files:
            _fileref = str(_file.relative_to(path))

            if _fileref not in _signatures:
                raise NotitiaPathExtensionUnreferencedFile("Missing signature for file")

            _contents = _file.read_text()
            _signature = hmac.new(
                self.signing_key.encode(), _contents.encode(), hashlib.sha256
            ).hexdigest()

            if _signature != _signatures[_fileref]:
                raise NotitiaPathExtensionFileSignatureInvalid(
                    f"Signature verification failed for {_fileref}"
                )

    @staticmethod
    def _import_module(name: str, source: Path) -> ModuleType:
        try:
            spec = importlib.util.spec_from_file_location(
                "notitia_module." + name.replace("/", "."), source
            )
            loader = importlib.util.LazyLoader(spec.loader)
            spec.loader = loader
            module = importlib.util.module_from_spec(spec)
            loader.exec_module(module)
            if hasattr(module, "BaseNotitiaModule"):
                return False

            if not hasattr(module, "NotitiaModule"):
                raise ValueError(
                    f"Invalid module {module} - NotitiaModule class not defined"
                )
            return module

        except Exception as e:
            print(f"failed to load module {source} // {str(e)}")
            return False



class NotitiaError(Exception):
    pass


class NotitiaBootstrapError(NotitiaError):
    pass


class NotitiaInvalidDriverException(NotitiaError):
    pass


class NotitiaInvalidDriverParameter(NotitiaError):
    pass


class NotitiaUnsupportedDriverCall(NotitiaError):
    pass


class NotitiaInvalidCasePayloadException(NotitiaError):
    pass


class NotitiaPathExtensionWithoutSignature(NotitiaError):
    pass


class NotitiaPathExtensionMissingSignature(NotitiaError):
    pass


class NotitiaPathExtensionUnreferencedFile(NotitiaError):
    pass


class NotitiaPathExtensionFileSignatureInvalid(NotitiaError):
    pass


class NotitiaUnknownModule(NotitiaError):
    pass


class NotitiaMalformedModulePath(NotitiaError):
    pass


class NotitiaMissingModule(NotitiaError):
    pass


class NotitiaInvalidModulePath(NotitiaError):
    pass


class NotitiaUnknownObserver(NotitiaError):
    pass



from __future__ import annotations
from types import ModuleType
import typing

from .roster import Roster, RosterType


if typing.TYPE_CHECKING:
    from lib.abc.investigator import InvestigatorABC
    from lib.abc.observer import ObserverABC
    from lib.abc.informant import InformantABC
    from lib.abc.reporter import ReporterABC

from .loader import Loader
from .error import (
    NotitiaBootstrapError,
    NotitiaInvalidDriverException,
)


class Notitia:
    __slots__ = (
        "driver",
        "driver_settings",
        "department",
        "observers",
        "informants",
        "reporters",
        "investigators",
        "_module_loader",
        "_supported_drivers",
        "_code_signature",
        "_roster",
    )

    driver: typing.AnyStr
    driver_settings: typing.Dict
    department: typing.Optional[typing.AnyStr]
    observers: typing.Dict[typing.AnyStr, ObserverABC]
    informants: typing.Dict[typing.AnyStr, InformantABC]
    reporters: typing.Dict[typing.AnyStr, ReporterABC]
    investigators: typing.Dict[typing.AnyStr, InvestigatorABC]
    _module_loader: Loader
    _supported_drivers: typing.Set[typing.AnyStr]
    _code_signature: typing.Optional[typing.AnyStr]
    _roster: typing.Optional[Roster]

    def __init__(
        self,
        department: typing.Optional[typing.AnyStr] = None,
        code_signature: typing.Optional[typing.AnyStr] = None,
    ):
        self.department = department

        self.driver = None
        self.driver_settings = None
        self._roster = None

        self.observers = self._module_loader.list_modules("observer")
        self.informants = self._module_loader.list_modules("informant")
        self.reporters = self._module_loader.list_modules("reporter")
        self.investigators = self._module_loader.list_modules("investigator")
        self._supported_drivers = self._module_loader.list_modules("driver")
        self._code_signature = code_signature

    def set_driver(
        self,
        driver: typing.AnyStr,
        driver_settings: typing.Dict,
    ):
        if driver not in self._supported_drivers:
            raise NotitiaInvalidDriverException()

        self.driver = driver
        self.driver_settings = driver_settings

    def get_from_registry(
        self,
        type: typing.AnyStr,
        key: typing.AnyStr,
    ) -> ModuleType[InvestigatorABC, ObserverABC, InformantABC, ReporterABC]:
        return self._module_loader.get_module(type, key)

    def extend(self, paths: typing.List[typing.AnyStr]):
        self._module_loader.signing_key = self._code_signature
        self._module_loader.add_paths(paths)

    def enlist(
        self,
        observer: typing.Optional[typing.AnyStr] = None,
        informant: typing.Optional[typing.AnyStr] = None,
        reporter: typing.Optional[typing.AnyStr] = None,
        investigator: typing.Optional[typing.AnyStr] = None,
    ):
        if observer:
            if observer not in self.observers:
                raise ValueError("Invalid observer")
            self._module_loader.load_module("observer", observer)

        if informant:
            if informant not in self.informants:
                raise ValueError("Invalid informant")
            self._module_loader.load_module("informant", informant)

        if reporter:
            if reporter not in self.reporters:
                raise ValueError("Invalid reporter")
            self._module_loader.load_module("reporter", reporter)

        if investigator:
            if investigator not in self.investigators:
                raise ValueError("Invalid investigator")
            self._module_loader.load_module("investigator", investigator)

    def clock_on(
        self,
        investigator: str,
        investigator_params: typing.Optional[typing.Dict] = None,
        dependency_params: typing.Optional[
            typing.Dict[str, typing.Dict[str, typing.Dict]]
        ] = None,
        *args,
        **kwargs,
    ):
        if investigator not in self.investigators:
            raise ValueError("Invalid investigator")

        _mod = self.get_from_registry("investigator", investigator)
        if _mod is False:
            raise ValueError("Investigator not enlisted or failed to load")

        _obj = self._load_and_resolve_investigator(
            _mod,
            investigator_params=investigator_params,
            dependency_params=dependency_params,
        )
        self._add_to_roster(
            _obj,
            {"args": args, "kwargs": kwargs},
        )

    def _boot_investigator(
        self, investigator_module, investigator_params
    ) -> InvestigatorABC:
        _obj: InvestigatorABC = investigator_module.NotitiaModule(
            department=self.department
        )
        _obj.sign_on(**investigator_params)
        return _obj

    def _load_and_resolve_investigator(
        self,
        investigator_module: ModuleType[InvestigatorABC],
        investigator_params: typing.Optional[typing.Dict] = None,
        dependency_params: typing.Optional[
            typing.Dict[str, typing.Dict[str, typing.Dict]]
        ] = None,
    ):
        if not investigator_params:
            investigator_params = {}

        _deps = getattr(investigator_module, "NOTITIA_DEPENDENCIES", None)
        if not _deps:
            return self._boot_investigator(investigator_module, investigator_params)

        _modules = self._resolve_module_dependencies(_deps)

        for _mod_type, _mod_entries in _deps.items():
            if not isinstance(_mod_entries, list):
                _mod_entries = [_mod_entries]

            for _mod_config in _mod_entries:
                _mod_key = f"{_mod_type}/{_mod_config['type']}"
                if _mod_key not in _modules or not _modules[_mod_key]["module"]:
                    if _mod_config.get("required", True):
                        raise RuntimeError(
                            f"Failed to load required dependency {_mod_key}"
                        )
                    continue

        for _type in ("driver", "observer", "informant", "reporter"):
            self._resolve_investigator_params(
                _type, investigator_params, dependency_params, _modules, _deps
            )

        return self._boot_investigator(investigator_module, investigator_params)

    def _resolve_investigator_params(  # noqa C90
        self,
        dep_type: str,
        investigator_params: typing.Dict[str, typing.Dict],
        dependency_params: typing.Dict[str, typing.Dict],
        modules: typing.Dict[str, typing.Dict],
        deps: typing.Dict[str, typing.Dict],
    ):
        # driver must come first, because other modules may request the driver
        if dep_type not in deps:
            return

        _entries = deps[dep_type]
        if not isinstance(_entries, list):
            _entries = [_entries]

        _di = {}
        for _entry in _entries:
            _mod = modules[f"{dep_type}/{_entry['type']}"]
            if (
                dependency_params
                and dep_type in dependency_params
                and _entry["type"] in dependency_params[dep_type]
            ):
                _mod["params"].update(dependency_params[dep_type][_entry["type"]])

            _req_drivers = []
            if (
                dep_type != "driver"
                and "driver" in investigator_params
                and investigator_params["driver"]
            ):
                _req_drivers = getattr(_mod["module"], "NOTITIA_DEPENDENCIES", {}).get(
                    "driver", []
                )
                if _req_drivers and not isinstance(_req_drivers, list):
                    _req_drivers = [_req_drivers]

            if _req_drivers:
                if "params" not in _mod:
                    _mod["params"] = {}
                if "driver" not in _mod["params"]:
                    _mod["params"]["driver"] = {}

            for _req_driver in _req_drivers:
                _t = _req_driver["type"]
                if _t not in investigator_params["driver"]:
                    continue

                _mod["params"]["driver"][_t] = investigator_params["driver"][_t]

            _di[_entry["type"]] = self._load_resolve_module(
                _mod["module"], init_params=_mod["params"]
            )

        investigator_params[dep_type] = _di

    def _resolve_module_dependencies(
        self,
        _deps: typing.Dict,
        _modules: typing.Optional[typing.Dict[str, typing.Dict]] = None,
    ) -> typing.Dict[str, typing.Dict]:
        is_nested = True
        if not _modules:
            is_nested = False
            _modules = {}

        # First pass - load modules
        for _mod_type, _mod_entries in _deps.items():
            if not isinstance(_mod_entries, list):
                _mod_entries = [_mod_entries]

            for _mod_config in _mod_entries:
                _mod_key = f"{_mod_type}/{_mod_config['type']}"
                if _mod_key in _modules:
                    continue

                _modules[_mod_key] = {
                    "module": self.get_from_registry(_mod_type, _mod_config["type"]),
                    "params": _mod_config.get("params", {}),
                }

                _subdeps = getattr(
                    _modules[f"{_mod_type}/{_mod_config['type']}"],
                    "NOTITIA_DEPENDENCIES",
                    None,
                )
                if _subdeps:
                    self._resolve_module_dependencies(_subdeps, _modules)

        if is_nested:
            return

        return _modules

    def _load_resolve_module(
        self, _mod: ModuleType, init_params: typing.Optional[typing.Dict]
    ):
        try:
            return _mod.NotitiaModule(**init_params, department=self.department)
        except TypeError as ex:
            raise NotitiaBootstrapError(_mod.__file__) from ex

    def _add_to_roster(
        self, investigator: InvestigatorABC, startup_params: typing.Dict
    ):
        if not self._roster:
            self._roster = Roster()

        self._roster.add(
            investigator=investigator,
            startup_params=startup_params,
            type=RosterType.PERMANENT,
        )

    def get_roster(self) -> Roster:
        if not self._roster:
            raise Exception("Roster is empty")

        return self._roster



from datetime import datetime, UTC, timezone
import dateparser
import typing
from bstk_notitia.lib.abc.informant import InformantABC

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "null",
        "required": True,
    },
}


class NotitiaModule(InformantABC):
    name = "Internal DateTime Informant"
    key = "internal/datetime"

    async def enquire(self, input: typing.Optional[str] = None):
        if not input:
            return datetime.now(tz=UTC)
        return self.parse(input)

    async def inform(self):
        pass

    def parse(self, date_string: str, tz: timezone = None) -> datetime:
        _date = dateparser.parse(date_string)

        if tz:
            return _date.astimezone(tz=tz)

        return _date



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import typing

if typing.TYPE_CHECKING:
    from lib.abc.driver import DriverABC

"""
A Notitia informant has a single job, which is to wait for an investigator
to ask for information, then do its best to get that information.
They never go off on their own to find new information and they always
ensure the information provided back to the investigator is in the
agreed format.
"""

NOTITIA_DEPENDENCIES: typing.Dict


@dataclass
class InformantABC(ABC):
    department: typing.AnyStr = field(kw_only=True, default=None)
    name: typing.AnyStr = field(init=False)
    key: typing.AnyStr = field(init=False)
    type: typing.AnyStr = field(init=False, default="informant")

    driver: typing.Dict[typing.AnyStr, DriverABC] = field(
        init=True, kw_only=True, default=None
    )

    def sign_on(self) -> typing.Dict:
        pass

    @abstractmethod
    async def enquire(self, **kwargs):
        pass

    @abstractmethod
    async def inform(self):
        pass



import typing

from bstk_notitia.lib.abc.driver import DriverABC


class NotitiaModule(DriverABC):
    name = "Null driver"
    key = "driver/null"
    connections: typing.Dict[str, None]

    # Notitia
    def connect(self, dsn: str) -> None:
        if dsn not in self.connections:
            self.connections[dsn] = None

        return self.connections[dsn]

    def disconnect(self, dsn: typing.Optional[str] = None):
        if dsn:
            if dsn not in self.connections:
                return

            del self.connections[dsn]
            return



import copy
from dataclasses import dataclass, field
import decimal
import typing


from bstk_notitia.lib.abc.driver import DriverABC

import bson
from bson import codec_options, Timestamp
import motor.motor_asyncio
import motor.docstrings
from pymongo import timeout
import pymongo.errors

from datetime import datetime, date, time, UTC
from enum import Enum


def _bson_encode(value):
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, decimal.Decimal):
        return bson.Decimal128(value)
    if (
        isinstance(value, object)
        and hasattr(value, "_to_json")
        and callable(value._to_json)
    ):
        return value._to_json()

    if isinstance(value, date):
        return datetime.combine(value, time.min)

    return None


class DecimalCodec(codec_options.TypeCodec):
    python_type = decimal.Decimal  # the Python type acted upon by this type codec
    bson_type = bson.Decimal128  # the BSON type acted upon by this type codec

    def transform_python(self, value: decimal.Decimal):
        """Function that transforms a custom type value into a type
        that BSON can encode. We use a precision of 30 to avoid overflows
        and other weird shit that happens during floatfuckery"""
        return bson.Decimal128(decimal.Context(prec=30).create_decimal(value))

    def transform_bson(self, value: bson.Decimal128):
        """Function that transforms a vanilla BSON type value into our
        custom type."""
        return value.to_decimal()


mongod_params = {
    "connectTimeoutMS": "2000",
    "heartbeatFrequencyMS": "1000",
    "serverSelectionTimeoutMS": "5000",
    "tz_aware": True,
    "uuidRepresentation": "pythonLegacy",
    "type_registry": codec_options.TypeRegistry(
        [DecimalCodec()], fallback_encoder=_bson_encode
    ),
}


@dataclass
class NotitiaModule(DriverABC):
    """
    The mongodb driver module provides a thin facade around mongo_motor.

    In general, all modules can use different collections and databases,
    but the DSN is managed per-department.

    Implementors can be specific, targeting databases and collections,
    or implicit, using the parameters provided by the investigator during
    startup.

    Regardless of which implementation is used, the only interface provided
    by this module that wouldn't normally be available is the `watch` command
    which provides a functional wrapper around a mongo change stream.
    """

    name = "MongoDB (Motor) driver"
    key = "driver/mongodb"

    dsn: typing.Optional[typing.AnyStr] = field(kw_only=True, default=None)
    params: typing.Optional[typing.Dict] = field(kw_only=True, default_factory=dict)
    database: typing.Optional[typing.AnyStr] = field(kw_only=True, default=None)
    collection: typing.Optional[typing.AnyStr] = field(kw_only=True, default=None)
    connections: typing.Dict[typing.AnyStr, motor.motor_asyncio.AsyncIOMotorClient] = (
        field(init=False, default_factory=dict)
    )

    def __post_init__(self):
        if True:
            # Fixed in 4.8 [https://jira.mongodb.org/browse/PYTHON-4449]
            def _change_stream_options(self) -> dict[str, typing.Any]:
                """Return the options dict for the $changeStream pipeline stage."""
                options: dict[str, typing.Any] = {}
                if self._full_document is not None:
                    options["fullDocument"] = self._full_document

                if self._full_document_before_change is not None:
                    options["fullDocumentBeforeChange"] = (
                        self._full_document_before_change
                    )

                resume_token = self.resume_token
                if resume_token is not None:
                    if self._uses_start_after:
                        options["startAfter"] = resume_token
                    else:
                        options["resumeAfter"] = resume_token

                elif self._start_at_operation_time is not None:
                    options["startAtOperationTime"] = self._start_at_operation_time

                if self._show_expanded_events:
                    options["showExpandedEvents"] = self._show_expanded_events

                return options

            from pymongo.change_stream import ChangeStream

            ChangeStream._change_stream_options = _change_stream_options

    # Notitia
    def connect(
        self, dsn: typing.Optional[typing.AnyStr] = None
    ) -> motor.motor_asyncio.AsyncIOMotorClient:
        if not dsn and self.dsn:
            dsn = self.dsn

        if dsn not in self.connections:
            _params = {**mongod_params, **self.params}
            self.connections[dsn] = motor.motor_asyncio.AsyncIOMotorClient(
                dsn, **_params
            )

        return self.connections[dsn]

    def disconnect(self, dsn: typing.Optional[str] = None):
        if dsn:
            if dsn not in self.connections:
                return

            self.connections[dsn].close()
            return

        for _conn in self.connections.values():
            _conn.close()

    # Native
    def get_db(
        self,
        name: typing.Optional[str] = None,
        client: typing.Optional[motor.motor_asyncio.AsyncIOMotorClient] = None,
    ) -> motor.motor_asyncio.AsyncIOMotorDatabase:
        if client is None and self.dsn:
            client = self.connect(self.dsn)

        if not name and self.database:
            name = self.database

        return client[name]

    def get_collection(
        self,
        name: typing.Optional[str] = None,
        db: typing.Optional[motor.motor_asyncio.AsyncIOMotorDatabase] = None,
    ) -> motor.motor_asyncio.AsyncIOMotorCollection:
        if db is None and self.dsn and self.database:
            db = self.get_db()

        if not name and self.collection:
            name = self.collection

        return db[name]

    def timestamp(self, date: typing.Optional[datetime] = None) -> Timestamp:
        if not date:
            date = datetime.now(UTC)
        return Timestamp(time=date, inc=0)

    def change_stream(
        self,
        pipeline: typing.List,
        collection: typing.Optional[motor.motor_asyncio.AsyncIOMotorCollection] = None,
        resume_token: typing.Optional[typing.AnyStr] = None,
        start_token: typing.Optional[typing.AnyStr] = None,
        **kwargs,
    ) -> motor.motor_asyncio.AsyncIOMotorChangeStream:
        if collection is None:
            if not self.collection:
                raise ValueError("Cannot auto-select collection")
            collection = self.get_collection(self.collection)

        params = {**kwargs}

        if not params:
            params = {}
        params["pipeline"] = pipeline
        if resume_token:
            params["resume_after"] = resume_token
        if start_token:
            params["start_after"] = start_token

        return collection.watch(**params)

    async def watch(
        self,
        pipeline: typing.Optional[typing.List[typing.Dict]] = None,
        collection: typing.Optional[motor.motor_asyncio.AsyncIOMotorCollection] = None,
        resume_token: typing.Optional[typing.AnyStr] = None,
        start_token: typing.Optional[typing.AnyStr] = None,
        changestream: typing.Optional[
            motor.motor_asyncio.AsyncIOMotorChangeStream
        ] = None,
        with_timeout: typing.Optional[int] = None,
        **kwargs,
    ) -> typing.AsyncGenerator[typing.Tuple[typing.AnyStr, typing.Dict], None]:

        if not changestream:
            if not pipeline:
                raise ValueError(
                    "If changestream is not provided, you must provide something for the pipeline"
                )

            changestream = self.change_stream(
                pipeline=pipeline,
                collection=collection,
                resume_token=resume_token,
                start_token=start_token,
                **kwargs,
            )

        while changestream:
            async with changestream as stream:
                while True:
                    try:
                        with timeout(with_timeout):
                            change = await stream.next()
                        resume_token = stream.resume_token
                        yield (resume_token, change)

                    except pymongo.errors.NetworkTimeout as ex:
                        if ex.timeout:
                            yield (resume_token, None)
                            continue

                    except pymongo.errors.OperationFailure as ex:
                        # If we tripped over an "invalidate", a 260 will be thrown.
                        # If there's a resume token within the changstream, use it as a start token
                        # and go around one last time
                        if ex.code == 260 and changestream.resume_token:
                            start_token = copy.deepcopy(changestream.resume_token)
                            resume_token = None
                        elif resume_token is None:
                            raise

                        changestream = self.change_stream(
                            pipeline=pipeline,
                            collection=collection,
                            resume_token=resume_token,
                            start_token=start_token,
                            **kwargs,
                        )
                        break

                    except Exception as ex:
                        raise ex



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timezone
import typing


if typing.TYPE_CHECKING:
    from bstk_notitia.lib.abc.driver import DriverABC
    from bstk_notitia.lib.abc.observer import ObserverABC
    from bstk_notitia.lib.abc.informant import InformantABC
    from bstk_notitia.lib.abc.reporter import ReporterABC


"""
Investigators specialise in working with a singular input,
collecting more information if required, deciding what to do
with that information, and then acting on that decision.

A good investigator has a narrow focus and works to quickly
close their case. They are a specialist at one particular task
and should not take more than a few seconds to come to a
conclusion (I/O not withstanding)

Complicated investigations lead to mistakes, so its usually
a good idea to have multiple investigators who specialise in
working with a discrete set of information and either create
new cases for other investigators, or reporting on their
case and closing it.

Abstracts aside, investigators:
  - hold the business logic
  - configure observers, informants and reporters
  - never change information directly
  - ensure each case is isolated from others
"""

NOTITIA_DEPENDENCIES: typing.Dict


@dataclass
class InvestigatorABC(ABC):
    busy: typing.Optional[bool] = field(init=False, default=None)
    department: typing.AnyStr = field(kw_only=True, default=None)
    name: typing.AnyStr = field(init=False)
    key: typing.AnyStr = field(init=False)
    type: typing.AnyStr = field(init=False, default="investigator")

    driver: typing.Dict[typing.AnyStr, DriverABC] = field(
        init=True, kw_only=True, default=None
    )
    observer: typing.Dict[typing.AnyStr, ObserverABC] = field(init=False, default=None)
    informant: typing.Dict[typing.AnyStr, InformantABC] = field(
        init=False, default=None
    )
    reporter: typing.Dict[typing.AnyStr, ReporterABC] = field(init=False, default=None)

    def sign_on(
        self,
        driver: typing.Optional[typing.Dict[typing.AnyStr, DriverABC]] = None,
        observer: typing.Optional[typing.Dict[typing.AnyStr, ObserverABC]] = None,
        informant: typing.Optional[typing.Dict[typing.AnyStr, InformantABC]] = None,
        reporter: typing.Optional[typing.Dict[typing.AnyStr, ReporterABC]] = None,
    ) -> typing.Dict:
        if driver:
            self.driver = driver
        if observer:
            self.observer = observer
        if informant:
            self.informant = informant
        if reporter:
            self.reporter = reporter

    @abstractmethod
    async def start(self, **kwargs):
        """
        'start' working... that could mean waiting for information from
        an observer, immediately reaching out to an informant, directly
        providing information to a report, or reach for a snack and wait..

        Investigators do whatever they do. An Investigator can flag whether they're
        "busy" or not (ie: interruptable) in an effort to accomodate graceful shutdowns.

        Without setting self.busy, the workstation will happily turn off at any time,
        regardless of what the investigator is doing. To prevent this, investiators should
        set `self.busy = True` when starting a stream of uninterruptable work and revert
        to `self.busy = False` when they're interruptable.

        A good example of when to use this is when working with a mongo collection stream..
        Without setting the busy flag, the workstation could and will stop the investigator
        mid-process, leaving a broken case behind and requiring cleanup.

        Setting this flag is the only way to ensure an investigator can complete case work
        cleanly during system maintenance events and deployments..
        """

        pass

    def working(self, state: bool):
        self.busy = state

    async def _stamp(
        self,
        document: typing.Dict,
        reporter: ReporterABC,
        updates: typing.Optional[typing.Dict] = None,
        constraints: typing.Optional[typing.Dict] = None,
    ):
        return await reporter.receive(
            action="update",
            record=document,
            data=updates,
            constraints=constraints,
        )

    async def mark(
        self,
        document: typing.Dict,
        reporter: ReporterABC,
        new_state: str,
        takeover: typing.Optional[bool] = False,
    ):

        _log = (
            f"{self.key} #{document['number']} ({document['category']}) state change "
        )
        _log += f"{document['state']} -> {new_state}"

        constraints = {
            "state": {"$eq": document["state"]},
        }
        if takeover is not True:
            constraints["assigned_investigator"] = {"$in": [None, self.key]}

        updates = {
            "state": new_state,
            "assigned_investigator": self.key,
        }
        if self.informant and "internal/datetime" in self.informant:
            updates["date"] = await self.informant["internal/datetime"].enquire()
        else:
            updates["date"] = datetime.now(tz=timezone.utc)

        _log += f" @ {updates['date'].isoformat()}"

        _stat = await self._stamp(
            document=document,
            reporter=reporter,
            updates=updates,
            constraints=constraints,
        )
        if _stat:
            document["state"] = new_state

        await self._log(_log, _stat)
        return _stat

    async def close(
        self,
        document: typing.Dict,
        reporter: ReporterABC,
        updates: typing.Optional[typing.Dict],
    ):
        return await self.record(
            document=document,
            reporter=reporter,
            updates=updates,
            new_state="complete",
        )

    async def record(
        self,
        document: typing.Dict,
        reporter: ReporterABC,
        updates: typing.Dict,
        new_state: typing.Optional[str] = None,
    ):
        _log = f"{self.key} #{document['number']} ({document['category']}) case update [{updates}] "
        _log += f"{document['state']} -> {new_state}"

        constraints = {
            "state": {"$eq": document["state"]},
            "assigned_investigator": {"$eq": self.key},
        }

        if new_state:
            updates["state"] = new_state
        if "internal/datetime" in self.informant:
            updates["date"] = await self.informant["internal/datetime"].enquire()
        else:
            updates["date"] = datetime.now(tz="UTC")

        _log += f" @ {updates['date'].isoformat()}"

        _stat = await self._stamp(
            document=document,
            reporter=reporter,
            updates=updates,
            constraints=constraints,
        )
        if _stat and new_state:
            document["state"] = new_state

        await self._log(_log, _stat)
        return _stat

    async def _log(self, message, stat):
        print(("✔︎ " if stat else "✕ ") + message)



from __future__ import annotations
import asyncio
from dataclasses import dataclass, field
import sys
import traceback
import typing
from bstk_notitia.lib.abc.investigator import InvestigatorABC

if typing.TYPE_CHECKING:
    import datetime
    from modules.driver.mongodb import NotitiaModule as MongoDBDriver, DriverABC

"""
This sample investigator is a mongodb record "enricher".

It asks the observer to collect records from a
particular collection that have not been collected,
gets the current date from the date reporter
and tells the mongo_record reporter that it
has changes to report.

Being a busywork investigator, it always tells
the reporter that it has a new record to report
before the observer is even engaged - so its
always got _something_ to do.
"""

NOTITIA_DEPENDENCIES = {
    "driver": {
        "type": "mongodb",
        "required": True,
        "params": {
            "dsn": "mongodb://localhost",
        },
    },
    "observer": {
        "type": "internal/mongo_collectionstream",
        "required": True,
        "params": {
            "lookup": {"reported": {"$ne": True}},
        },
    },
    "informant": {
        "type": "internal/datetime",
        "required": True,
    },
    "reporter": [
        {
            "type": "internal/mongo_record",
            "required": True,
        },
        {
            "type": "internal/echo",
            "required": True,
        },
    ],
}


@dataclass
class NotitiaModule(InvestigatorABC):
    name = "Internal Notitia Case Reporter"
    key = "internal/notitia_casereporter"
    driver: typing.Dict[str, typing.Union[MongoDBDriver, DriverABC]]

    case_limit: typing.SupportsInt = field(init=False, default=None)

    def sign_on(
        self,
        case_limit: typing.Optional[typing.SupportsInt] = None,
        *args,
        **kwargs,
    ):
        self.case_limit = case_limit
        super().sign_on(*args, **kwargs)

    async def start(self, **kwargs):
        if not self.case_limit:
            self.case_limit = 10

        try:
            case_counter = 0
            while self.case_limit == 0 or case_counter < self.case_limit:
                print("tick..")
                async for document in self.observer[
                    "internal/mongo_collectionstream"
                ].monitor():
                    case_counter += 1
                    data: datetime.datetime = await self.informant[
                        "internal/datetime"
                    ].enquire()

                    res = await self.reporter["internal/mongo_record"].receive(
                        action="update",
                        record=document,
                        data={"reported": True, "reported_at": data},
                    )
                    await self.reporter["internal/echo"].receive(
                        {
                            f"Reporting on case {case_counter} of {self.case_limit} @ {data.isoformat()}": document,
                            "Result": res,
                        }
                    )
                    if self.case_limit and case_counter >= self.case_limit:
                        break

                await asyncio.sleep(1)

        except Exception:
            traceback.print_exc(file=sys.stdout, chain=True)
            raise

        return
